-module(asn1ct_parser).
-define(THIS_MODULE, asn1ct_parser).
-export([parse/1, parse_and_scan/1, format_error/1]).
%%-author('kenneth@erix.ericsson.se').
-copyright('Copyright (c) 1991-99 Ericsson Telecom AB').
-vsn('$Revision: /main/release/1 $').
-include("asn1_records.hrl").

to_set(V) when list(V) ->
	ordsets:list_to_set(V);
to_set(V) ->
	ordsets:list_to_set([V]).

merge_constraints({Rlist,ExtList}) -> % extensionmarker in constraint
	{merge_constraints(Rlist,[],[]),
	 merge_constraints(ExtList,[],[])};
	
merge_constraints(Clist) ->
	merge_constraints(Clist, [], []).

merge_constraints([Ch|Ct],Cacc, Eacc) ->
	NewEacc = case Ch#constraint.e of
			undefined -> Eacc;
			E -> [E|Eacc]
		  end,
	merge_constraints(Ct,[fixup_constraint(Ch#constraint.c)|Cacc],NewEacc);

merge_constraints([],Cacc,[]) ->
	lists:flatten(Cacc);
merge_constraints([],Cacc,Eacc) ->
	lists:flatten(Cacc) ++ [{'Errors',Eacc}].

fixup_constraint(C) ->
	case C of	
		{'SingleValue',V} when list(V) ->
			[C, 
			{'ValueRange',{lists:min(V),lists:max(V)}}]; 
		{'PermittedAlphabet',{'SingleValue',V}} when list(V) ->
			V2 = {'SingleValue',
			       ordsets:list_to_set(lists:flatten(V))},
			{'PermittedAlphabet',V2};
		{'PermittedAlphabet',{'SingleValue',V}} ->
			V2 = {'SingleValue',[V]},
			{'PermittedAlphabet',V2};
		{'SizeConstraint',{'SingleValue',V}} ->
			NewV = case V of
				[Sv] ->
					Sv;
				L when list(L) ->
					ordsets:list_to_set(V);
				Else ->
					Else
			       end,
			{'SizeConstraint',NewV};
		{'SizeConstraint',{'ValueRange',{Lb,Lb}}} ->
			{'SizeConstraint',Lb};
		{'SizeConstraint',{'ValueRange',{Lb,Ub}}} ->
			{'SizeConstraint',{Lb,Ub}};
		List when list(List) ->
			[fixup_constraint(Xc)||Xc <- List]; 	
		Other ->
			Other
	end.












%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?THIS_MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?THIS_MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?THIS_MODULE, "end_of_file"}};
        {error, Descriptor, Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?THIS_MODULE, Reason}};
        {ok, Tokens, Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?THIS_MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?THIS_MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({Cat, _, Val}) -> io_lib:format('~w', [Val]);

yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


yeccpars2(0, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 3, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(1, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(1, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(2, 'DEFINITIONS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [2 | __Ss], [__T | __Stack]);
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 5, [3 | __Ss], [__T | __Stack]);
yeccpars2(3, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(4, __Cat, [3 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(4, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   put(asn1_module,__1#typereference.val), {__1,__2}
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ModuleIdentifier', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(5, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [6 | __Ss], [__T | __Stack]);
yeccpars2(6, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [6 | __Ss], [__T | __Stack]);
yeccpars2(6, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('DefinitiveObjIdComponentList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(7, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [7 | __Ss], [__T | __Stack]);
yeccpars2(7, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(8, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [8 | __Ss], [__T | __Stack]);
yeccpars2(8, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('DefinitiveObjIdComponent', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(9, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('DefinitiveObjIdComponent', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(10, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [10 | __Ss], [__T | __Stack]);
yeccpars2(10, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(11, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [11 | __Ss], [__T | __Stack]);
yeccpars2(11, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(12, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('DefinitiveObjIdComponent', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(13, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('DefinitiveIdentifier', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(14, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('DefinitiveObjIdComponentList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(15, 'AUTOMATIC', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [15 | __Ss], [__T | __Stack]);
yeccpars2(15, 'IMPLICIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [15 | __Ss], [__T | __Stack]);
yeccpars2(15, 'EXPLICIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [15 | __Ss], [__T | __Stack]);
yeccpars2(15, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val = begin
   put(tagdefault,'EXPLICIT'), 'EXPLICIT'
  end,
 yeccpars2(19, __Cat, [15 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(16, 'TAGS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 435, [16 | __Ss], [__T | __Stack]);
yeccpars2(16, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(17, 'TAGS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 434, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(18, 'TAGS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 433, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(19, 'EXTENSIBILITY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [19 | __Ss], [__T | __Stack]);
yeccpars2(19, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  false,
 yeccpars2(21, __Cat, [19 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(20, 'IMPLIED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 432, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(21, '::=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [21 | __Ss], [__T | __Stack]);
yeccpars2(21, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(22, 'BEGIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [22 | __Ss], [__T | __Stack]);
yeccpars2(22, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(23, 'EXPORTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, 'END', __Ss, __Stack, __T, __Ts, __Tzr) ->
 __Val =  {'ModuleBody',nil,nil,[]},
 yeccpars2(26, 'END', [23 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(23, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  {exports,all},
 yeccpars2(25, __Cat, [23 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(24, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 413, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 412, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 429, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(25, 'IMPORTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  {imports,[]},
 yeccpars2(29, __Cat, [25 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(26, 'END', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [26 | __Ss], [__T | __Stack]);
yeccpars2(26, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(27, __Cat, __Ss,  [__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {'ModuleBody',Ex,Im,Types} = __7, {{typereference,Pos,Name},Defid} = __1, #module{pos = Pos, name = Name, defid = Defid, tagdefault = __3, extensiondefault = __4, exports = Ex, imports = Im, typeorval = Types}
  end,
 __Nss = lists:nthtail(7, __Ss),
 yeccpars2(yeccgoto('ModuleDefinition', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(28, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 413, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 412, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 405, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(29, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [29 | __Ss], [__T | __Stack]);
yeccpars2(29, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [29 | __Ss], [__T | __Stack]);
yeccpars2(29, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(30, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('AssignmentList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ModuleBody',__1,__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ModuleBody', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(32, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Assignment', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(33, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Assignment', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(34, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Assignment', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(35, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Assignment', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(36, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(37, '::=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(38, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, 'CLASS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 361, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(39, '::=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(40, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(41, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('Parameters', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Parameter', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ParameterList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(45, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [45 | __Ss], [__T | __Stack]);
yeccpars2(45, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(46, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Parameters', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(47, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(48, 'DEFINED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 358, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(357, __Cat, [48 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(49, 'STRING', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 353, [49 | __Ss], [__T | __Stack]);
yeccpars2(49, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(50, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(51, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = __1},
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(53, 'STRING', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 352, [53 | __Ss], [__T | __Stack]);
yeccpars2(53, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(54, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 349, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ReferencedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(59, 'PDV', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 348, [59 | __Ss], [__T | __Stack]);
yeccpars2(59, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(60, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 339, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(61, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(62, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(63, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('DefinedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(64, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'GeneralizedTime',
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(65, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 331, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'INTEGER',
 yeccpars2(yeccgoto('IntegerType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(66, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = 'NULL'},
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, 'IDENTIFIER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 330, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(69, 'STRING', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 329, [69 | __Ss], [__T | __Stack]);
yeccpars2(69, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(70, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'ObjectDescriptor',
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(71, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(72, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = __1},
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [73 | __Ss], [__T | __Stack]);
yeccpars2(73, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 318, [73 | __Ss], [__T | __Stack]);
yeccpars2(73, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 317, [73 | __Ss], [__T | __Stack]);
yeccpars2(73, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 319, [73 | __Ss], [__T | __Stack]);
yeccpars2(73, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(74, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 290, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(75, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ReferencedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(76, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(77, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(78, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(79, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(80, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'EXPLICIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 283, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'IMPLICIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 284, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(81, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Type', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(82, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #ptypedef{pos = element(2,__1), name = element(3,__1), args = __2, typespec = __4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('ParameterizedTypeAssignment', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(83, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ReferencedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(84, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ConstrainedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'UTCTime',
 yeccpars2(yeccgoto('BuiltinType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(86, 'PRIVATE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 271, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'APPLICATION', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 269, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'UNIVERSAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 272, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  'CONTEXT',
 yeccpars2(270, __Cat, [86 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(87, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(88, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('CharacterStringType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(89, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'Externaltypereference'{pos = __1#typereference.pos, module = get(asn1_module), type = __1#typereference.val},
 yeccpars2(yeccgoto('DefinedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(90, typefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 93, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(91, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {pt,__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('DefinedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(92, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TypeFromObject', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(93, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'Externaltypereference'{pos = element(2,__1), module = element(3,__1), type = element(3,__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Externaltypereference', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(94, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(95, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SelectionType',element(3,__1),__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SelectionType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(96, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'ALL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 103, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'CONSTRAINED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(97, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1#type{constraint = merge_constraints([__2])},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ConstrainedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(98, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1#type{constraint = merge_constraints([__2,__3])},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ConstrainedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(99, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1#type{constraint = merge_constraints([__2,__3,__4])},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('ConstrainedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(100, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1#type{constraint = merge_constraints([__2,__3,__4,__5])},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('ConstrainedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(101, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'ALL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(102, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 266, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(103, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 264, [103 | __Ss], [__T | __Stack]);
yeccpars2(103, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(104, 'EXCEPT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 224, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(105, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(106, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ValueNotNull', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(107, 'BY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 243, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(108, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(109, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(110, '!', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 231, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  undefined,
 yeccpars2(232, __Cat, [110 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(111, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 227, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ElementSetSpecs', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(112, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ConstraintSpec', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(113, 'EXCEPT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('Elems', hd(__Ss)), 'EXCEPT', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(113, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('IntersectionElements', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(114, 'EXCEPT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 224, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(115, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  false,
 yeccpars2(yeccgoto('BooleanValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(116, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(117, '^', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 221, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'INTERSECTION', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 219, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(118, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(119, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Intersections', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(120, 'INTERSECTION', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('IElems', hd(__Ss)), 'INTERSECTION', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(120, '^', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('IElems', hd(__Ss)), '^', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(120, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Unions', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(121, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 217, [121 | __Ss], [__T | __Stack]);
yeccpars2(121, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('LowerEndpoint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(122, '..', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 210, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(123, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'MIN',
 yeccpars2(yeccgoto('LowerEndValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(124, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('Value', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(125, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('ValueList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(126, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('RestrictedCharacterStringValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(127, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Elements', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(128, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('CharacterStringValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(129, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(130, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(131, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(132, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Elements', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(133, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  true,
 yeccpars2(yeccgoto('BooleanValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('RestrictedCharacterStringValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(135, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 206, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, 'UNION', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 204, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(136, 'UNION', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('UElems', hd(__Ss)), 'UNION', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(136, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('UElems', hd(__Ss)), '|', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(136, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ElementSetSpec', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(137, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ConstraintSpec', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(138, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 178, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, '..', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('LowerEndValue', hd(__Ss)), '..', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(138, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('LowerEndValue', hd(__Ss)), '<', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(138, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('ValueList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(139, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SingleValue',__1},
 yeccpars2(yeccgoto('SubTypeElements', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(140, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Value', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(141, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 183, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, 'COMPONENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 182, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(142, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(143, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('RestrictedCharacterStringValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(144, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('BuiltinValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(145, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ValueNotNull', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(146, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('SignedNumber', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(147, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 181, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'Externaltypereference'{pos = __1#typereference.pos, module = get(asn1_module), type = __1#typereference.val},
 yeccpars2(yeccgoto('DefinedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(148, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(149, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 178, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('ValueList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(150, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(151, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ValueNotNull', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(152, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 157, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('SignedNumber', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(153, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 155, [153 | __Ss], [__T | __Stack]);
yeccpars2(153, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(154, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SAndSOfValue', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(155, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 156, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(156, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'Externalvaluereference'{pos = element(2,__1), module = element(3,__1), value = element(3,__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ValueNotNull', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(157, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 158, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(158, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 160, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 159, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(159, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(160, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'Tuple',__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('Tuple', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(161, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 162, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(162, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(163, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [163 | __Ss], [__T | __Stack]);
yeccpars2(163, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(164, __Cat, __Ss,  [__9,__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'Quadruple',__2,__4,__6,__8},
 __Nss = lists:nthtail(8, __Ss),
 yeccpars2(yeccgoto('Quadruple', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(165, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(166, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(167, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ChoiceValue',element(3,__1),__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ChoiceValue', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(168, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ValueNotNull', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(169, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 176, [169 | __Ss], [__T | __Stack]);
yeccpars2(169, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(170, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 175, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(171, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 172, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(172, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 173, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(173, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(174, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedNumber',element(3,__1),{'ExternalValue',element(3,__3),element(3,__5)}},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('NamedNumber', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(175, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedNumber',element(3,__1),element(3,__3)},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('NamedNumber', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(176, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedNumber',element(3,__1),__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('NamedNumber', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(177, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SAndSOfValue', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(178, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(179, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1,space|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ValueList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(180, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ValueList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(181, typefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [181 | __Ss], [__T | __Stack]);
yeccpars2(181, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 156, [181 | __Ss], [__T | __Stack]);
yeccpars2(181, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 93, [181 | __Ss], [__T | __Stack]);
yeccpars2(181, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(182, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(183, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 184, [183 | __Ss], [__T | __Stack]);
yeccpars2(183, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(184, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 188, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 185, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(185, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 200, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(186, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 198, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('TypeConstraints', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(187, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(188, 'OPTIONAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 192, [188 | __Ss], [__T | __Stack]);
yeccpars2(188, 'ABSENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 190, [188 | __Ss], [__T | __Stack]);
yeccpars2(188, 'PRESENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 193, [188 | __Ss], [__T | __Stack]);
yeccpars2(188, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 189, [188 | __Ss], [__T | __Stack]);
yeccpars2(188, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [188 | __Ss], [__T | __Stack]);
yeccpars2(188, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('TypeConstraints', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 188, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(190, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'ABSENT',
 yeccpars2(yeccgoto('PresenceConstraint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(191, 'OPTIONAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 192, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ABSENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 190, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'PRESENT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 193, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedConstraint',element(3,__1),__2,undefined},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('NamedConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(192, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'OPTIONAL',
 yeccpars2(yeccgoto('PresenceConstraint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(193, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'PRESENT',
 yeccpars2(yeccgoto('PresenceConstraint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(194, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedConstraint',element(3,__1),undefined,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('NamedConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(195, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'NamedConstraint',element(3,__1),__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('NamedConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(196, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TypeConstraints', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'WITH COMPONENTS',{'FullSpecification',__4}},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(198, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 188, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(199, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TypeConstraints', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(200, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 188, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(201, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 202, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(202, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'WITH COMPONENTS',{'PartialSpecification',__3}},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(203, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'WITH COMPONENT',__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UnionMark', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(205, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(206, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UnionMark', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(207, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  case {__1,__3} of {{'SingleValue',V1},{'SingleValue',V2}} -> {'SingleValue',ordsets:union(to_set(V1),to_set(V2))} end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Unions', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(208, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SizeConstraint',__2#constraint.c},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(209, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1,space|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ValueList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(210, 'MAX', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(211, 'MAX', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(212, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'MAX',
 yeccpars2(yeccgoto('UpperEndValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(213, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UpperEndpoint', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(214, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ValueRange',{__1,__3}},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UpperEndValue', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(216, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 - 1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('UpperEndpoint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(217, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1 + 1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('LowerEndpoint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(218, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ContainedSubType',__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(219, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('IntersectionMark', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(220, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(221, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('IntersectionMark', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(222, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  case {__1,__3} of {{'SingleValue',V1},{'SingleValue',V2}} -> {'SingleValue',ordsets:intersection(to_set(V1),to_set(V2))}; {V1,V2} when list(V1) -> V1 ++ [V2]; {V1,V2} -> [V1,V2] end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Intersections', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(223, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'PermittedAlphabet',__2#constraint.c},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SubTypeElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(224, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(225, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('IntersectionElements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(226, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'EXCEPT',__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('Exclusions', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(227, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 228, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(228, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,[]},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ElementSetSpecs', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(229, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ALL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(230, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('ElementSetSpecs', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(231, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 238, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 237, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(232, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 233, [232 | __Ss], [__T | __Stack]);
yeccpars2(232, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(233, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #constraint{c = __2, e = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('Constraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(234, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ExceptionSpec', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(235, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ExceptionIdentification', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(236, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 241, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(237, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ExceptionIdentification', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(238, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 239, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'Externaltypereference'{pos = __1#typereference.pos, module = get(asn1_module), type = __1#typereference.val},
 yeccpars2(yeccgoto('DefinedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(239, typefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 240, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 93, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(240, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'Externalvaluereference'{pos = element(2,__1), module = element(3,__1), value = element(3,__1)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ExceptionIdentification', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(241, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(242, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ExceptionIdentification', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(243, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 244, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(244, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 253, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(245, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('UserDefinedConstraintParameter', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(246, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 260, [246 | __Ss], [__T | __Stack]);
yeccpars2(246, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [246 | __Ss], [__T | __Stack]);
yeccpars2(246, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ActualParameter', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(247, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 258, [247 | __Ss], [__T | __Stack]);
yeccpars2(247, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('UserDefinedConstraintParameters', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(248, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 257, [248 | __Ss], [__T | __Stack]);
yeccpars2(248, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(249, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ActualParameter', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(250, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ActualParameter', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(251, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ValueNotNull', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(252, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'ALL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(253, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {constrained_by,[]},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('UserDefinedConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 256, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(255, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SingleValue',__1},
 yeccpars2(yeccgoto('SubTypeElements', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(256, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {valueset,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ValueSet', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(257, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {constrained_by,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('UserDefinedConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(258, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(259, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('UserDefinedConstraintParameters', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(260, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(261, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('UserDefinedConstraintParameter', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(262, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [262 | __Ss], [__T | __Stack]);
yeccpars2(262, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ActualParameter', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(263, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ALL',__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ElementSetSpec', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(264, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'MIN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'SIZE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'INCLUDES', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ALL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(265, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {[],__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ElementSetSpecs', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(266, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  - element(3,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SignedNumber', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(267, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 268, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(268, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Elements', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(269, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('Class', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(270, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 274, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 275, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 276, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(271, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('Class', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(272, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto('Class', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(273, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 282, [273 | __Ss], [__T | __Stack]);
yeccpars2(273, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(274, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 281, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('ClassNumber', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(275, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 280, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('ClassNumber', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(276, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 277, [276 | __Ss], [__T | __Stack]);
yeccpars2(276, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(277, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 278, [277 | __Ss], [__T | __Stack]);
yeccpars2(277, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(278, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 279, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(279, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #tag{class = __2, number = #'Externalvaluereference'{pos = element(2,__3), module = element(3,__3), value = element(3,__5)}},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('Tag', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(280, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #tag{class = __2, number = element(3,__3)},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('Tag', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(281, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #tag{class = __2, number = element(3,__3)},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('Tag', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(282, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #tag{class = __2, number = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('Tag', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(283, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(284, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(285, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [285 | __Ss], [__T | __Stack]);
yeccpars2(285, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2#type{tag = __1#tag{type = {default,get(tagdefault)}}},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('TaggedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(286, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [286 | __Ss], [__T | __Stack]);
yeccpars2(286, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3#type{tag = __1#tag{type = 'IMPLICIT'}},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TaggedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(287, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3#type{tag = __1#tag{type = 'EXPLICIT'}},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TaggedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(288, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 314, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(289, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(290, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(291, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 298, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 299, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 293, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 292, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(292, '!', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 231, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  undefined,
 yeccpars2(309, __Cat, [292 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(293, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 307, [293 | __Ss], [__T | __Stack]);
yeccpars2(293, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(294, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 305, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('ComponentTypeList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(295, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(296, 'DEFAULT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 301, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, 'OPTIONAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {'NamedType',Pos,{Name,Type}} = __1, #'ComponentType'{pos = Pos, name = Name, typespec = Type, prop = mandatory}
  end,
 yeccpars2(yeccgoto('ComponentType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(297, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('NamedType', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(298, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(299, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SET',[]},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SetType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(300, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [300 | __Ss], [__T | __Stack]);
yeccpars2(300, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   V1 = __1, {'NamedType',V1#identifier.pos,{V1#identifier.val,__2}}
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('NamedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(301, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(302, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {'NamedType',Pos,{Name,Type}} = __1, #'ComponentType'{pos = Pos, name = Name, typespec = Type, prop = 'OPTIONAL'}
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ComponentType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(303, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {'NamedType',Pos,{Name,Type}} = __1, #'ComponentType'{pos = Pos, name = Name, typespec = Type, prop = {'DEFAULT',__3}}
  end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ComponentType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(304, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SET',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('SetType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(305, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 298, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 293, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 292, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(306, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ComponentTypeList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(307, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(308, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [308 | __Ss], [__T | __Stack]);
yeccpars2(308, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'COMPONENTS OF',__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ComponentType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(309, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'EXTENSIONMARK',element(2,__1),__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ComponentType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(310, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 311, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(311, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(312, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = {'SET OF',__5}, constraint = merge_constraints([#constraint{c = {'SizeConstraint',__3#constraint.c}}])},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('TypeWithConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(313, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [313 | __Ss], [__T | __Stack]);
yeccpars2(313, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SET OF',__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SetOfType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(314, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(315, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [315 | __Ss], [__T | __Stack]);
yeccpars2(315, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = {'SET OF',__4}, constraint = merge_constraints([__2])},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('TypeWithConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(316, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 327, [316 | __Ss], [__T | __Stack]);
yeccpars2(316, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(317, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(318, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [318 | __Ss], [__T | __Stack]);
yeccpars2(318, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(319, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 298, [319 | __Ss], [__T | __Stack]);
yeccpars2(319, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 293, [319 | __Ss], [__T | __Stack]);
yeccpars2(319, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 292, [319 | __Ss], [__T | __Stack]);
yeccpars2(319, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 321, [319 | __Ss], [__T | __Stack]);
yeccpars2(319, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(320, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 322, [320 | __Ss], [__T | __Stack]);
yeccpars2(320, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(321, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SEQUENCE',[]},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SequenceType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(322, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SEQUENCE',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('SequenceType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(323, 'OF', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [323 | __Ss], [__T | __Stack]);
yeccpars2(323, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(324, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(325, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [325 | __Ss], [__T | __Stack]);
yeccpars2(325, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = {'SEQUENCE OF',__5}, constraint = merge_constraints([#constraint{c = {'SizeConstraint',__3#constraint.c}}])},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('TypeWithConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(326, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [326 | __Ss], [__T | __Stack]);
yeccpars2(326, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'SEQUENCE OF',__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SequenceOfType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(327, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(328, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [328 | __Ss], [__T | __Stack]);
yeccpars2(328, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #type{def = {'SEQUENCE OF',__4}, constraint = merge_constraints([__2])},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('TypeWithConstraint', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(329, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'OCTET STRING',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('BuiltinType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(330, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'OBJECT IDENTIFIER',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('BuiltinType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(331, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 334, [331 | __Ss], [__T | __Stack]);
yeccpars2(331, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(332, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 336, [332 | __Ss], [__T | __Stack]);
yeccpars2(332, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('NamedNumberList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(333, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 335, [333 | __Ss], [__T | __Stack]);
yeccpars2(333, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(334, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [334 | __Ss], [__T | __Stack]);
yeccpars2(334, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(335, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'INTEGER',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('IntegerType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(336, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 334, [336 | __Ss], [__T | __Stack]);
yeccpars2(336, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(337, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('NamedNumberList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(338, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {pt,__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('DefinedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(339, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 340, [339 | __Ss], [__T | __Stack]);
yeccpars2(339, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 344, [339 | __Ss], [__T | __Stack]);
yeccpars2(339, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(340, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'EXTENSIONMARK',
 yeccpars2(yeccgoto('EnumerationItem', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(341, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 347, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(342, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 345, [342 | __Ss], [__T | __Stack]);
yeccpars2(342, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('Enumeration', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(343, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('EnumerationItem', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(344, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [344 | __Ss], [__T | __Stack]);
yeccpars2(344, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto('EnumerationItem', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(345, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 340, [345 | __Ss], [__T | __Stack]);
yeccpars2(345, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 344, [345 | __Ss], [__T | __Stack]);
yeccpars2(345, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(346, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Enumeration', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(347, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'ENUMERATED',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('EnumeratedType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(348, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'EMBEDDED PDV',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('BuiltinType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(349, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 298, [349 | __Ss], [__T | __Stack]);
yeccpars2(349, 'COMPONENTS', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 293, [349 | __Ss], [__T | __Stack]);
yeccpars2(349, '...', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 292, [349 | __Ss], [__T | __Stack]);
yeccpars2(349, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(350, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 351, [350 | __Ss], [__T | __Stack]);
yeccpars2(350, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(351, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'CHOICE',__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('ChoiceType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(352, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'CHARACTER STRING',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('CharacterStringType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(353, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 354, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'BIT STRING',[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('BitStringType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(354, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 334, [354 | __Ss], [__T | __Stack]);
yeccpars2(354, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(355, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 356, [355 | __Ss], [__T | __Stack]);
yeccpars2(355, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(356, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'BIT STRING',__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('BitStringType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(357, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  case __2 of [] -> 'ANY'; _ -> {'ANY DEFINED BY',__2} end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('BuiltinType', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(358, 'BY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 359, [358 | __Ss], [__T | __Stack]);
yeccpars2(358, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(359, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 360, [359 | __Ss], [__T | __Stack]);
yeccpars2(359, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(360, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('AnyDefBy', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(361, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 363, [361 | __Ss], [__T | __Stack]);
yeccpars2(361, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(362, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #typedef{pos = element(2,__1), name = element(3,__1), typespec = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('TypeAssignment', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(363, valuefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 367, [363 | __Ss], [__T | __Stack]);
yeccpars2(363, typefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 366, [363 | __Ss], [__T | __Stack]);
yeccpars2(363, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(364, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 399, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('FieldSpecs', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(365, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 376, [365 | __Ss], [__T | __Stack]);
yeccpars2(365, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(366, 'OPTIONAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 371, [366 | __Ss], [__T | __Stack]);
yeccpars2(366, 'DEFAULT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 370, [366 | __Ss], [__T | __Stack]);
yeccpars2(366, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  'MANDATORY',
 yeccpars2(375, __Cat, [366 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(367, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(368, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [368 | __Ss], [__T | __Stack]);
yeccpars2(368, 'UNIQUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 369, [368 | __Ss], [__T | __Stack]);
yeccpars2(368, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(369, 'OPTIONAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 371, [369 | __Ss], [__T | __Stack]);
yeccpars2(369, 'DEFAULT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 370, [369 | __Ss], [__T | __Stack]);
yeccpars2(369, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  'MANDATORY',
 yeccpars2(372, __Cat, [369 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(370, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'SEQUENCE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'SET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'CHARACTER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, restrictedcharacterstringtype, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'CHOICE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'BIT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'ENUMERATED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'INTEGER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'ObjectDescriptor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'UTCTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'GeneralizedTime', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'REAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'OCTET', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'OBJECT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'EXTERNAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'EMBEDDED', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'BOOLEAN', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'ANY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [370 | __Ss], [__T | __Stack]);
yeccpars2(370, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(371, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'OPTIONAL',
 yeccpars2(yeccgoto('OptionalitySpec', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(372, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {valuefield,__1,__2,'UNIQUE',__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('FieldSpec', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(373, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [373 | __Ss], [__T | __Stack]);
yeccpars2(373, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'DEFAULT',__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OptionalitySpec', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(374, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  case __2 of {identifier,_,Id} -> {'DEFAULT',Id}; _ -> {'DEFAULT',__2} end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OptionalitySpec', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(375, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {typefield,__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('FieldSpec', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(376, 'WITH', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 377, [376 | __Ss], [__T | __Stack]);
yeccpars2(376, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #typedef{pos = element(2,__1), name = element(3,__1), typespec = {'CLASS',__5,[]}},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('ObjectClassAssignment', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(377, 'SYNTAX', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 379, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(378, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #typedef{pos = element(2,__1), name = element(3,__1), typespec = {'CLASS',__5,__7}},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto('ObjectClassAssignment', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(379, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 381, [379 | __Ss], [__T | __Stack]);
yeccpars2(379, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(380, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'WITH SYNTAX',__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('WithSyntaxSpec', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(381, valuefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 393, [381 | __Ss], [__T | __Stack]);
yeccpars2(381, typefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 391, [381 | __Ss], [__T | __Stack]);
yeccpars2(381, 'BY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 383, [381 | __Ss], [__T | __Stack]);
yeccpars2(381, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 382, [381 | __Ss], [__T | __Stack]);
yeccpars2(381, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 392, [381 | __Ss], [__T | __Stack]);
yeccpars2(381, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 390, [381 | __Ss], [__T | __Stack]);
yeccpars2(381, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 394, [381 | __Ss], [__T | __Stack]);
yeccpars2(381, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(382, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('RequiredToken', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(383, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'BY',
 yeccpars2(yeccgoto('Word', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(384, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('TokenOrGroupSpec', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(385, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('RequiredToken', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(386, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('TokenOrGroupSpec', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(387, valuefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 393, [387 | __Ss], [__T | __Stack]);
yeccpars2(387, typefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 391, [387 | __Ss], [__T | __Stack]);
yeccpars2(387, 'BY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 383, [387 | __Ss], [__T | __Stack]);
yeccpars2(387, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 382, [387 | __Ss], [__T | __Stack]);
yeccpars2(387, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 392, [387 | __Ss], [__T | __Stack]);
yeccpars2(387, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 390, [387 | __Ss], [__T | __Stack]);
yeccpars2(387, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('TokenOrGroupSpecs', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(388, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 397, [388 | __Ss], [__T | __Stack]);
yeccpars2(388, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(389, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('RequiredToken', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(390, valuefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 393, [390 | __Ss], [__T | __Stack]);
yeccpars2(390, typefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 391, [390 | __Ss], [__T | __Stack]);
yeccpars2(390, 'BY', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 383, [390 | __Ss], [__T | __Stack]);
yeccpars2(390, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 382, [390 | __Ss], [__T | __Stack]);
yeccpars2(390, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 392, [390 | __Ss], [__T | __Stack]);
yeccpars2(390, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 390, [390 | __Ss], [__T | __Stack]);
yeccpars2(390, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(391, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('PrimitiveFieldName', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(392, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('RequiredToken', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(393, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('PrimitiveFieldName', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(394, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SyntaxList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(395, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 396, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(396, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('OptionalGroup', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(397, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SyntaxList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(398, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('TokenOrGroupSpecs', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(399, valuefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 367, [399 | __Ss], [__T | __Stack]);
yeccpars2(399, typefieldreference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 366, [399 | __Ss], [__T | __Stack]);
yeccpars2(399, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(400, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('FieldSpecs', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(401, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [401 | __Ss], [__T | __Stack]);
yeccpars2(401, '::=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 402, [401 | __Ss], [__T | __Stack]);
yeccpars2(401, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(402, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(403, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #valuedef{pos = element(2,__1), name = element(3,__1), type = __2, value = __4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('ValueAssignment', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(404, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('AssignmentList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(405, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {imports,[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('Imports', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(406, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 427, [406 | __Ss], [__T | __Stack]);
yeccpars2(406, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(407, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Symbol', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(408, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 425, [408 | __Ss], [__T | __Stack]);
yeccpars2(408, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('SymbolList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(409, 'FROM', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 419, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(410, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 413, [410 | __Ss], [__T | __Stack]);
yeccpars2(410, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 412, [410 | __Ss], [__T | __Stack]);
yeccpars2(410, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('SymbolsFromModuleList', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(411, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 417, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(412, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Reference', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(413, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 414, [413 | __Ss], [__T | __Stack]);
yeccpars2(413, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 415, [413 | __Ss], [__T | __Stack]);
yeccpars2(413, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Reference', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(414, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 93, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(415, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 416, [415 | __Ss], [__T | __Stack]);
yeccpars2(415, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(416, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Reference', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(417, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {imports,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Imports', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(418, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('SymbolsFromModuleList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(419, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 420, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(420, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 421, [420 | __Ss], [__T | __Stack]);
yeccpars2(420, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 422, [420 | __Ss], [__T | __Stack]);
yeccpars2(420, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SymbolsFromModule'{symbols = __1, module = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SymbolsFromModule', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(421, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SymbolsFromModule'{symbols = __1, module = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('SymbolsFromModule', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(422, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, cstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, number, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, hstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, bstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, 'NULL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [422 | __Ss], [__T | __Stack]);
yeccpars2(422, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(423, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 424, [423 | __Ss], [__T | __Stack]);
yeccpars2(423, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(424, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SymbolsFromModule'{symbols = __1, module = __3},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('SymbolsFromModule', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(425, typereference, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 413, [425 | __Ss], [__T | __Stack]);
yeccpars2(425, identifier, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 412, [425 | __Ss], [__T | __Stack]);
yeccpars2(425, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(426, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('SymbolList', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(427, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 428, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(428, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Reference', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(429, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {exports,[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('Exports', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(430, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 431, [430 | __Ss], [__T | __Stack]);
yeccpars2(430, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(431, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {exports,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('Exports', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(432, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'IMPLIED',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ExtensionDefault', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(433, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   put(tagdefault,'IMPLICIT'), 'IMPLICIT'
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('TagDefault', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(434, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   put(tagdefault,'EXPLICIT'), 'EXPLICIT'
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('TagDefault', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(435, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   put(tagdefault,'AUTOMATIC'), 'AUTOMATIC'
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('TagDefault', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto('ActualParameter', 244) ->
 245;
yeccgoto('ActualParameter', 258) ->
 245;
yeccgoto('ActualParameter', 260) ->
 261;
yeccgoto('AnyDefBy', 48) ->
 357;
yeccgoto('Assignment', 29) ->
 30;
yeccgoto('Assignment', 30) ->
 30;
yeccgoto('AssignmentList', 29) ->
 31;
yeccgoto('AssignmentList', 30) ->
 404;
yeccgoto('BitStringType', 36) ->
 51;
yeccgoto('BitStringType', 38) ->
 51;
yeccgoto('BitStringType', 47) ->
 51;
yeccgoto('BitStringType', 80) ->
 51;
yeccgoto('BitStringType', 94) ->
 51;
yeccgoto('BitStringType', 118) ->
 51;
yeccgoto('BitStringType', 231) ->
 51;
yeccgoto('BitStringType', 244) ->
 51;
yeccgoto('BitStringType', 258) ->
 51;
yeccgoto('BitStringType', 260) ->
 51;
yeccgoto('BitStringType', 283) ->
 51;
yeccgoto('BitStringType', 284) ->
 51;
yeccgoto('BitStringType', 289) ->
 51;
yeccgoto('BitStringType', 298) ->
 51;
yeccgoto('BitStringType', 307) ->
 51;
yeccgoto('BitStringType', 311) ->
 51;
yeccgoto('BitStringType', 314) ->
 51;
yeccgoto('BitStringType', 317) ->
 51;
yeccgoto('BitStringType', 324) ->
 51;
yeccgoto('BitStringType', 327) ->
 51;
yeccgoto('BitStringType', 367) ->
 51;
yeccgoto('BitStringType', 370) ->
 51;
yeccgoto('BooleanValue', 96) ->
 105;
yeccgoto('BooleanValue', 101) ->
 105;
yeccgoto('BooleanValue', 125) ->
 105;
yeccgoto('BooleanValue', 138) ->
 105;
yeccgoto('BooleanValue', 148) ->
 105;
yeccgoto('BooleanValue', 149) ->
 105;
yeccgoto('BooleanValue', 166) ->
 105;
yeccgoto('BooleanValue', 178) ->
 105;
yeccgoto('BooleanValue', 205) ->
 105;
yeccgoto('BooleanValue', 210) ->
 105;
yeccgoto('BooleanValue', 211) ->
 105;
yeccgoto('BooleanValue', 220) ->
 105;
yeccgoto('BooleanValue', 224) ->
 105;
yeccgoto('BooleanValue', 229) ->
 105;
yeccgoto('BooleanValue', 241) ->
 105;
yeccgoto('BooleanValue', 244) ->
 105;
yeccgoto('BooleanValue', 252) ->
 105;
yeccgoto('BooleanValue', 258) ->
 105;
yeccgoto('BooleanValue', 260) ->
 105;
yeccgoto('BooleanValue', 264) ->
 105;
yeccgoto('BooleanValue', 301) ->
 105;
yeccgoto('BooleanValue', 370) ->
 105;
yeccgoto('BooleanValue', 402) ->
 105;
yeccgoto('BooleanValue', 422) ->
 105;
yeccgoto('BuiltinType', 36) ->
 52;
yeccgoto('BuiltinType', 38) ->
 52;
yeccgoto('BuiltinType', 47) ->
 52;
yeccgoto('BuiltinType', 80) ->
 52;
yeccgoto('BuiltinType', 94) ->
 52;
yeccgoto('BuiltinType', 118) ->
 52;
yeccgoto('BuiltinType', 231) ->
 52;
yeccgoto('BuiltinType', 244) ->
 52;
yeccgoto('BuiltinType', 258) ->
 52;
yeccgoto('BuiltinType', 260) ->
 52;
yeccgoto('BuiltinType', 283) ->
 52;
yeccgoto('BuiltinType', 284) ->
 52;
yeccgoto('BuiltinType', 289) ->
 52;
yeccgoto('BuiltinType', 298) ->
 52;
yeccgoto('BuiltinType', 307) ->
 52;
yeccgoto('BuiltinType', 311) ->
 52;
yeccgoto('BuiltinType', 314) ->
 52;
yeccgoto('BuiltinType', 317) ->
 52;
yeccgoto('BuiltinType', 324) ->
 52;
yeccgoto('BuiltinType', 327) ->
 52;
yeccgoto('BuiltinType', 367) ->
 52;
yeccgoto('BuiltinType', 370) ->
 52;
yeccgoto('BuiltinValue', 96) ->
 106;
yeccgoto('BuiltinValue', 101) ->
 106;
yeccgoto('BuiltinValue', 125) ->
 106;
yeccgoto('BuiltinValue', 138) ->
 106;
yeccgoto('BuiltinValue', 148) ->
 106;
yeccgoto('BuiltinValue', 149) ->
 106;
yeccgoto('BuiltinValue', 166) ->
 106;
yeccgoto('BuiltinValue', 178) ->
 106;
yeccgoto('BuiltinValue', 205) ->
 106;
yeccgoto('BuiltinValue', 210) ->
 106;
yeccgoto('BuiltinValue', 211) ->
 106;
yeccgoto('BuiltinValue', 220) ->
 106;
yeccgoto('BuiltinValue', 224) ->
 106;
yeccgoto('BuiltinValue', 229) ->
 106;
yeccgoto('BuiltinValue', 241) ->
 106;
yeccgoto('BuiltinValue', 244) ->
 106;
yeccgoto('BuiltinValue', 252) ->
 106;
yeccgoto('BuiltinValue', 258) ->
 106;
yeccgoto('BuiltinValue', 260) ->
 106;
yeccgoto('BuiltinValue', 264) ->
 106;
yeccgoto('BuiltinValue', 301) ->
 106;
yeccgoto('BuiltinValue', 370) ->
 106;
yeccgoto('BuiltinValue', 402) ->
 106;
yeccgoto('BuiltinValue', 422) ->
 106;
yeccgoto('CharacterStringType', 36) ->
 55;
yeccgoto('CharacterStringType', 38) ->
 55;
yeccgoto('CharacterStringType', 47) ->
 55;
yeccgoto('CharacterStringType', 80) ->
 55;
yeccgoto('CharacterStringType', 94) ->
 55;
yeccgoto('CharacterStringType', 118) ->
 55;
yeccgoto('CharacterStringType', 231) ->
 55;
yeccgoto('CharacterStringType', 244) ->
 55;
yeccgoto('CharacterStringType', 258) ->
 55;
yeccgoto('CharacterStringType', 260) ->
 55;
yeccgoto('CharacterStringType', 283) ->
 55;
yeccgoto('CharacterStringType', 284) ->
 55;
yeccgoto('CharacterStringType', 289) ->
 55;
yeccgoto('CharacterStringType', 298) ->
 55;
yeccgoto('CharacterStringType', 307) ->
 55;
yeccgoto('CharacterStringType', 311) ->
 55;
yeccgoto('CharacterStringType', 314) ->
 55;
yeccgoto('CharacterStringType', 317) ->
 55;
yeccgoto('CharacterStringType', 324) ->
 55;
yeccgoto('CharacterStringType', 327) ->
 55;
yeccgoto('CharacterStringType', 367) ->
 55;
yeccgoto('CharacterStringType', 370) ->
 55;
yeccgoto('CharacterStringValue', 96) ->
 108;
yeccgoto('CharacterStringValue', 101) ->
 108;
yeccgoto('CharacterStringValue', 125) ->
 108;
yeccgoto('CharacterStringValue', 138) ->
 108;
yeccgoto('CharacterStringValue', 148) ->
 108;
yeccgoto('CharacterStringValue', 149) ->
 108;
yeccgoto('CharacterStringValue', 166) ->
 108;
yeccgoto('CharacterStringValue', 178) ->
 108;
yeccgoto('CharacterStringValue', 205) ->
 108;
yeccgoto('CharacterStringValue', 210) ->
 108;
yeccgoto('CharacterStringValue', 211) ->
 108;
yeccgoto('CharacterStringValue', 220) ->
 108;
yeccgoto('CharacterStringValue', 224) ->
 108;
yeccgoto('CharacterStringValue', 229) ->
 108;
yeccgoto('CharacterStringValue', 241) ->
 108;
yeccgoto('CharacterStringValue', 244) ->
 108;
yeccgoto('CharacterStringValue', 252) ->
 108;
yeccgoto('CharacterStringValue', 258) ->
 108;
yeccgoto('CharacterStringValue', 260) ->
 108;
yeccgoto('CharacterStringValue', 264) ->
 108;
yeccgoto('CharacterStringValue', 301) ->
 108;
yeccgoto('CharacterStringValue', 370) ->
 108;
yeccgoto('CharacterStringValue', 402) ->
 108;
yeccgoto('CharacterStringValue', 422) ->
 108;
yeccgoto('ChoiceType', 36) ->
 56;
yeccgoto('ChoiceType', 38) ->
 56;
yeccgoto('ChoiceType', 47) ->
 56;
yeccgoto('ChoiceType', 80) ->
 56;
yeccgoto('ChoiceType', 94) ->
 56;
yeccgoto('ChoiceType', 118) ->
 56;
yeccgoto('ChoiceType', 231) ->
 56;
yeccgoto('ChoiceType', 244) ->
 56;
yeccgoto('ChoiceType', 258) ->
 56;
yeccgoto('ChoiceType', 260) ->
 56;
yeccgoto('ChoiceType', 283) ->
 56;
yeccgoto('ChoiceType', 284) ->
 56;
yeccgoto('ChoiceType', 289) ->
 56;
yeccgoto('ChoiceType', 298) ->
 56;
yeccgoto('ChoiceType', 307) ->
 56;
yeccgoto('ChoiceType', 311) ->
 56;
yeccgoto('ChoiceType', 314) ->
 56;
yeccgoto('ChoiceType', 317) ->
 56;
yeccgoto('ChoiceType', 324) ->
 56;
yeccgoto('ChoiceType', 327) ->
 56;
yeccgoto('ChoiceType', 367) ->
 56;
yeccgoto('ChoiceType', 370) ->
 56;
yeccgoto('ChoiceValue', 96) ->
 109;
yeccgoto('ChoiceValue', 101) ->
 109;
yeccgoto('ChoiceValue', 125) ->
 109;
yeccgoto('ChoiceValue', 138) ->
 109;
yeccgoto('ChoiceValue', 148) ->
 109;
yeccgoto('ChoiceValue', 149) ->
 109;
yeccgoto('ChoiceValue', 166) ->
 109;
yeccgoto('ChoiceValue', 178) ->
 109;
yeccgoto('ChoiceValue', 205) ->
 109;
yeccgoto('ChoiceValue', 210) ->
 109;
yeccgoto('ChoiceValue', 211) ->
 109;
yeccgoto('ChoiceValue', 220) ->
 109;
yeccgoto('ChoiceValue', 224) ->
 109;
yeccgoto('ChoiceValue', 229) ->
 109;
yeccgoto('ChoiceValue', 241) ->
 109;
yeccgoto('ChoiceValue', 244) ->
 109;
yeccgoto('ChoiceValue', 252) ->
 109;
yeccgoto('ChoiceValue', 258) ->
 109;
yeccgoto('ChoiceValue', 260) ->
 109;
yeccgoto('ChoiceValue', 264) ->
 109;
yeccgoto('ChoiceValue', 301) ->
 109;
yeccgoto('ChoiceValue', 370) ->
 109;
yeccgoto('ChoiceValue', 402) ->
 109;
yeccgoto('ChoiceValue', 422) ->
 109;
yeccgoto('Class', 86) ->
 270;
yeccgoto('ClassNumber', 270) ->
 273;
yeccgoto('ComponentType', 291) ->
 294;
yeccgoto('ComponentType', 305) ->
 294;
yeccgoto('ComponentType', 319) ->
 294;
yeccgoto('ComponentType', 349) ->
 294;
yeccgoto('ComponentTypeList', 291) ->
 295;
yeccgoto('ComponentTypeList', 305) ->
 306;
yeccgoto('ComponentTypeList', 319) ->
 320;
yeccgoto('ComponentTypeList', 349) ->
 350;
yeccgoto('ConstrainedType', 36) ->
 57;
yeccgoto('ConstrainedType', 38) ->
 57;
yeccgoto('ConstrainedType', 47) ->
 57;
yeccgoto('ConstrainedType', 80) ->
 57;
yeccgoto('ConstrainedType', 94) ->
 57;
yeccgoto('ConstrainedType', 118) ->
 57;
yeccgoto('ConstrainedType', 231) ->
 57;
yeccgoto('ConstrainedType', 244) ->
 57;
yeccgoto('ConstrainedType', 258) ->
 57;
yeccgoto('ConstrainedType', 260) ->
 57;
yeccgoto('ConstrainedType', 283) ->
 57;
yeccgoto('ConstrainedType', 284) ->
 57;
yeccgoto('ConstrainedType', 289) ->
 57;
yeccgoto('ConstrainedType', 298) ->
 57;
yeccgoto('ConstrainedType', 307) ->
 57;
yeccgoto('ConstrainedType', 311) ->
 57;
yeccgoto('ConstrainedType', 314) ->
 57;
yeccgoto('ConstrainedType', 317) ->
 57;
yeccgoto('ConstrainedType', 324) ->
 57;
yeccgoto('ConstrainedType', 327) ->
 57;
yeccgoto('ConstrainedType', 367) ->
 57;
yeccgoto('ConstrainedType', 370) ->
 57;
yeccgoto('Constraint', 73) ->
 316;
yeccgoto('Constraint', 74) ->
 288;
yeccgoto('Constraint', 82) ->
 97;
yeccgoto('Constraint', 95) ->
 97;
yeccgoto('Constraint', 97) ->
 98;
yeccgoto('Constraint', 98) ->
 99;
yeccgoto('Constraint', 99) ->
 100;
yeccgoto('Constraint', 116) ->
 223;
yeccgoto('Constraint', 130) ->
 208;
yeccgoto('Constraint', 182) ->
 203;
yeccgoto('Constraint', 188) ->
 191;
yeccgoto('Constraint', 218) ->
 97;
yeccgoto('Constraint', 236) ->
 97;
yeccgoto('Constraint', 246) ->
 97;
yeccgoto('Constraint', 262) ->
 97;
yeccgoto('Constraint', 285) ->
 97;
yeccgoto('Constraint', 286) ->
 97;
yeccgoto('Constraint', 287) ->
 97;
yeccgoto('Constraint', 290) ->
 310;
yeccgoto('Constraint', 300) ->
 97;
yeccgoto('Constraint', 308) ->
 97;
yeccgoto('Constraint', 312) ->
 97;
yeccgoto('Constraint', 313) ->
 97;
yeccgoto('Constraint', 315) ->
 97;
yeccgoto('Constraint', 318) ->
 323;
yeccgoto('Constraint', 325) ->
 97;
yeccgoto('Constraint', 326) ->
 97;
yeccgoto('Constraint', 328) ->
 97;
yeccgoto('Constraint', 362) ->
 97;
yeccgoto('Constraint', 368) ->
 97;
yeccgoto('Constraint', 373) ->
 97;
yeccgoto('Constraint', 401) ->
 97;
yeccgoto('ConstraintSpec', 96) ->
 110;
yeccgoto('DefinedType', 36) ->
 58;
yeccgoto('DefinedType', 38) ->
 58;
yeccgoto('DefinedType', 47) ->
 58;
yeccgoto('DefinedType', 80) ->
 58;
yeccgoto('DefinedType', 94) ->
 58;
yeccgoto('DefinedType', 96) ->
 58;
yeccgoto('DefinedType', 101) ->
 58;
yeccgoto('DefinedType', 118) ->
 58;
yeccgoto('DefinedType', 205) ->
 58;
yeccgoto('DefinedType', 220) ->
 58;
yeccgoto('DefinedType', 224) ->
 58;
yeccgoto('DefinedType', 229) ->
 58;
yeccgoto('DefinedType', 231) ->
 58;
yeccgoto('DefinedType', 244) ->
 58;
yeccgoto('DefinedType', 252) ->
 58;
yeccgoto('DefinedType', 258) ->
 58;
yeccgoto('DefinedType', 260) ->
 58;
yeccgoto('DefinedType', 264) ->
 58;
yeccgoto('DefinedType', 283) ->
 58;
yeccgoto('DefinedType', 284) ->
 58;
yeccgoto('DefinedType', 289) ->
 58;
yeccgoto('DefinedType', 298) ->
 58;
yeccgoto('DefinedType', 307) ->
 58;
yeccgoto('DefinedType', 311) ->
 58;
yeccgoto('DefinedType', 314) ->
 58;
yeccgoto('DefinedType', 317) ->
 58;
yeccgoto('DefinedType', 324) ->
 58;
yeccgoto('DefinedType', 327) ->
 58;
yeccgoto('DefinedType', 367) ->
 58;
yeccgoto('DefinedType', 370) ->
 58;
yeccgoto('DefinitiveIdentifier', 3) ->
 4;
yeccgoto('DefinitiveObjIdComponent', 5) ->
 6;
yeccgoto('DefinitiveObjIdComponent', 6) ->
 6;
yeccgoto('DefinitiveObjIdComponentList', 5) ->
 7;
yeccgoto('DefinitiveObjIdComponentList', 6) ->
 14;
yeccgoto('ElementSetSpec', 96) ->
 111;
yeccgoto('ElementSetSpec', 101) ->
 267;
yeccgoto('ElementSetSpec', 229) ->
 230;
yeccgoto('ElementSetSpec', 252) ->
 254;
yeccgoto('ElementSetSpec', 264) ->
 265;
yeccgoto('ElementSetSpecs', 96) ->
 112;
yeccgoto('Elements', 96) ->
 113;
yeccgoto('Elements', 101) ->
 113;
yeccgoto('Elements', 205) ->
 113;
yeccgoto('Elements', 220) ->
 113;
yeccgoto('Elements', 224) ->
 226;
yeccgoto('Elements', 229) ->
 113;
yeccgoto('Elements', 252) ->
 113;
yeccgoto('Elements', 264) ->
 113;
yeccgoto('Elems', 96) ->
 114;
yeccgoto('Elems', 101) ->
 114;
yeccgoto('Elems', 205) ->
 114;
yeccgoto('Elems', 220) ->
 114;
yeccgoto('Elems', 229) ->
 114;
yeccgoto('Elems', 252) ->
 114;
yeccgoto('Elems', 264) ->
 114;
yeccgoto('EnumeratedType', 36) ->
 62;
yeccgoto('EnumeratedType', 38) ->
 62;
yeccgoto('EnumeratedType', 47) ->
 62;
yeccgoto('EnumeratedType', 80) ->
 62;
yeccgoto('EnumeratedType', 94) ->
 62;
yeccgoto('EnumeratedType', 118) ->
 62;
yeccgoto('EnumeratedType', 231) ->
 62;
yeccgoto('EnumeratedType', 244) ->
 62;
yeccgoto('EnumeratedType', 258) ->
 62;
yeccgoto('EnumeratedType', 260) ->
 62;
yeccgoto('EnumeratedType', 283) ->
 62;
yeccgoto('EnumeratedType', 284) ->
 62;
yeccgoto('EnumeratedType', 289) ->
 62;
yeccgoto('EnumeratedType', 298) ->
 62;
yeccgoto('EnumeratedType', 307) ->
 62;
yeccgoto('EnumeratedType', 311) ->
 62;
yeccgoto('EnumeratedType', 314) ->
 62;
yeccgoto('EnumeratedType', 317) ->
 62;
yeccgoto('EnumeratedType', 324) ->
 62;
yeccgoto('EnumeratedType', 327) ->
 62;
yeccgoto('EnumeratedType', 367) ->
 62;
yeccgoto('EnumeratedType', 370) ->
 62;
yeccgoto('Enumeration', 339) ->
 341;
yeccgoto('Enumeration', 345) ->
 346;
yeccgoto('EnumerationItem', 339) ->
 342;
yeccgoto('EnumerationItem', 345) ->
 342;
yeccgoto('ExceptionIdentification', 231) ->
 234;
yeccgoto('ExceptionSpec', 110) ->
 232;
yeccgoto('ExceptionSpec', 292) ->
 309;
yeccgoto('Exclusions', 104) ->
 263;
yeccgoto('Exclusions', 114) ->
 225;
yeccgoto('Exports', 23) ->
 25;
yeccgoto('ExtensionDefault', 19) ->
 21;
yeccgoto('Externaltypereference', 24) ->
 406;
yeccgoto('Externaltypereference', 28) ->
 406;
yeccgoto('Externaltypereference', 36) ->
 63;
yeccgoto('Externaltypereference', 38) ->
 63;
yeccgoto('Externaltypereference', 47) ->
 63;
yeccgoto('Externaltypereference', 80) ->
 63;
yeccgoto('Externaltypereference', 94) ->
 63;
yeccgoto('Externaltypereference', 96) ->
 63;
yeccgoto('Externaltypereference', 101) ->
 63;
yeccgoto('Externaltypereference', 118) ->
 63;
yeccgoto('Externaltypereference', 205) ->
 63;
yeccgoto('Externaltypereference', 220) ->
 63;
yeccgoto('Externaltypereference', 224) ->
 63;
yeccgoto('Externaltypereference', 229) ->
 63;
yeccgoto('Externaltypereference', 231) ->
 63;
yeccgoto('Externaltypereference', 244) ->
 63;
yeccgoto('Externaltypereference', 252) ->
 63;
yeccgoto('Externaltypereference', 258) ->
 63;
yeccgoto('Externaltypereference', 260) ->
 63;
yeccgoto('Externaltypereference', 264) ->
 63;
yeccgoto('Externaltypereference', 283) ->
 63;
yeccgoto('Externaltypereference', 284) ->
 63;
yeccgoto('Externaltypereference', 289) ->
 63;
yeccgoto('Externaltypereference', 298) ->
 63;
yeccgoto('Externaltypereference', 307) ->
 63;
yeccgoto('Externaltypereference', 311) ->
 63;
yeccgoto('Externaltypereference', 314) ->
 63;
yeccgoto('Externaltypereference', 317) ->
 63;
yeccgoto('Externaltypereference', 324) ->
 63;
yeccgoto('Externaltypereference', 327) ->
 63;
yeccgoto('Externaltypereference', 367) ->
 63;
yeccgoto('Externaltypereference', 370) ->
 63;
yeccgoto('Externaltypereference', 410) ->
 406;
yeccgoto('Externaltypereference', 425) ->
 406;
yeccgoto('FieldSpec', 363) ->
 364;
yeccgoto('FieldSpec', 399) ->
 364;
yeccgoto('FieldSpecs', 363) ->
 365;
yeccgoto('FieldSpecs', 399) ->
 400;
yeccgoto('IElems', 96) ->
 117;
yeccgoto('IElems', 101) ->
 117;
yeccgoto('IElems', 229) ->
 117;
yeccgoto('IElems', 252) ->
 117;
yeccgoto('IElems', 264) ->
 117;
yeccgoto('Imports', 25) ->
 29;
yeccgoto('IntegerType', 36) ->
 66;
yeccgoto('IntegerType', 38) ->
 66;
yeccgoto('IntegerType', 47) ->
 66;
yeccgoto('IntegerType', 80) ->
 66;
yeccgoto('IntegerType', 94) ->
 66;
yeccgoto('IntegerType', 118) ->
 66;
yeccgoto('IntegerType', 231) ->
 66;
yeccgoto('IntegerType', 244) ->
 66;
yeccgoto('IntegerType', 258) ->
 66;
yeccgoto('IntegerType', 260) ->
 66;
yeccgoto('IntegerType', 283) ->
 66;
yeccgoto('IntegerType', 284) ->
 66;
yeccgoto('IntegerType', 289) ->
 66;
yeccgoto('IntegerType', 298) ->
 66;
yeccgoto('IntegerType', 307) ->
 66;
yeccgoto('IntegerType', 311) ->
 66;
yeccgoto('IntegerType', 314) ->
 66;
yeccgoto('IntegerType', 317) ->
 66;
yeccgoto('IntegerType', 324) ->
 66;
yeccgoto('IntegerType', 327) ->
 66;
yeccgoto('IntegerType', 367) ->
 66;
yeccgoto('IntegerType', 370) ->
 66;
yeccgoto('IntersectionElements', 96) ->
 119;
yeccgoto('IntersectionElements', 101) ->
 119;
yeccgoto('IntersectionElements', 205) ->
 207;
yeccgoto('IntersectionElements', 220) ->
 222;
yeccgoto('IntersectionElements', 229) ->
 119;
yeccgoto('IntersectionElements', 252) ->
 119;
yeccgoto('IntersectionElements', 264) ->
 119;
yeccgoto('IntersectionMark', 117) ->
 220;
yeccgoto('Intersections', 96) ->
 120;
yeccgoto('Intersections', 101) ->
 120;
yeccgoto('Intersections', 229) ->
 120;
yeccgoto('Intersections', 252) ->
 120;
yeccgoto('Intersections', 264) ->
 120;
yeccgoto('LowerEndValue', 96) ->
 121;
yeccgoto('LowerEndValue', 101) ->
 121;
yeccgoto('LowerEndValue', 205) ->
 121;
yeccgoto('LowerEndValue', 220) ->
 121;
yeccgoto('LowerEndValue', 224) ->
 121;
yeccgoto('LowerEndValue', 229) ->
 121;
yeccgoto('LowerEndValue', 252) ->
 121;
yeccgoto('LowerEndValue', 264) ->
 121;
yeccgoto('LowerEndpoint', 96) ->
 122;
yeccgoto('LowerEndpoint', 101) ->
 122;
yeccgoto('LowerEndpoint', 205) ->
 122;
yeccgoto('LowerEndpoint', 220) ->
 122;
yeccgoto('LowerEndpoint', 224) ->
 122;
yeccgoto('LowerEndpoint', 229) ->
 122;
yeccgoto('LowerEndpoint', 252) ->
 122;
yeccgoto('LowerEndpoint', 264) ->
 122;
yeccgoto('ModuleBody', 23) ->
 26;
yeccgoto('ModuleDefinition', 0) ->
 1;
yeccgoto('ModuleIdentifier', 0) ->
 2;
yeccgoto('NamedConstraint', 184) ->
 186;
yeccgoto('NamedConstraint', 189) ->
 186;
yeccgoto('NamedConstraint', 198) ->
 186;
yeccgoto('NamedConstraint', 200) ->
 186;
yeccgoto('NamedNumber', 96) ->
 125;
yeccgoto('NamedNumber', 101) ->
 125;
yeccgoto('NamedNumber', 125) ->
 125;
yeccgoto('NamedNumber', 138) ->
 125;
yeccgoto('NamedNumber', 148) ->
 125;
yeccgoto('NamedNumber', 149) ->
 125;
yeccgoto('NamedNumber', 178) ->
 125;
yeccgoto('NamedNumber', 205) ->
 125;
yeccgoto('NamedNumber', 220) ->
 125;
yeccgoto('NamedNumber', 224) ->
 125;
yeccgoto('NamedNumber', 229) ->
 125;
yeccgoto('NamedNumber', 252) ->
 125;
yeccgoto('NamedNumber', 264) ->
 125;
yeccgoto('NamedNumber', 331) ->
 332;
yeccgoto('NamedNumber', 336) ->
 332;
yeccgoto('NamedNumber', 339) ->
 343;
yeccgoto('NamedNumber', 345) ->
 343;
yeccgoto('NamedNumber', 354) ->
 332;
yeccgoto('NamedNumber', 422) ->
 125;
yeccgoto('NamedNumberList', 331) ->
 333;
yeccgoto('NamedNumberList', 336) ->
 337;
yeccgoto('NamedNumberList', 354) ->
 355;
yeccgoto('NamedType', 291) ->
 296;
yeccgoto('NamedType', 305) ->
 296;
yeccgoto('NamedType', 319) ->
 296;
yeccgoto('NamedType', 349) ->
 296;
yeccgoto('ObjectClassAssignment', 29) ->
 32;
yeccgoto('ObjectClassAssignment', 30) ->
 32;
yeccgoto('OptionalGroup', 381) ->
 384;
yeccgoto('OptionalGroup', 387) ->
 384;
yeccgoto('OptionalGroup', 390) ->
 384;
yeccgoto('OptionalitySpec', 366) ->
 375;
yeccgoto('OptionalitySpec', 369) ->
 372;
yeccgoto('Parameter', 40) ->
 41;
yeccgoto('Parameter', 45) ->
 41;
yeccgoto('ParameterList', 37) ->
 39;
yeccgoto('ParameterList', 63) ->
 338;
yeccgoto('ParameterList', 89) ->
 91;
yeccgoto('ParameterList', 147) ->
 91;
yeccgoto('ParameterList', 238) ->
 91;
yeccgoto('ParameterizedTypeAssignment', 29) ->
 33;
yeccgoto('ParameterizedTypeAssignment', 30) ->
 33;
yeccgoto('Parameters', 40) ->
 42;
yeccgoto('Parameters', 45) ->
 46;
yeccgoto('PresenceConstraint', 188) ->
 194;
yeccgoto('PresenceConstraint', 191) ->
 195;
yeccgoto('PrimitiveFieldName', 381) ->
 385;
yeccgoto('PrimitiveFieldName', 387) ->
 385;
yeccgoto('PrimitiveFieldName', 390) ->
 385;
yeccgoto('Quadruple', 96) ->
 126;
yeccgoto('Quadruple', 101) ->
 126;
yeccgoto('Quadruple', 125) ->
 126;
yeccgoto('Quadruple', 138) ->
 126;
yeccgoto('Quadruple', 148) ->
 126;
yeccgoto('Quadruple', 149) ->
 126;
yeccgoto('Quadruple', 166) ->
 126;
yeccgoto('Quadruple', 178) ->
 126;
yeccgoto('Quadruple', 205) ->
 126;
yeccgoto('Quadruple', 210) ->
 126;
yeccgoto('Quadruple', 211) ->
 126;
yeccgoto('Quadruple', 220) ->
 126;
yeccgoto('Quadruple', 224) ->
 126;
yeccgoto('Quadruple', 229) ->
 126;
yeccgoto('Quadruple', 241) ->
 126;
yeccgoto('Quadruple', 244) ->
 126;
yeccgoto('Quadruple', 252) ->
 126;
yeccgoto('Quadruple', 258) ->
 126;
yeccgoto('Quadruple', 260) ->
 126;
yeccgoto('Quadruple', 264) ->
 126;
yeccgoto('Quadruple', 301) ->
 126;
yeccgoto('Quadruple', 370) ->
 126;
yeccgoto('Quadruple', 402) ->
 126;
yeccgoto('Quadruple', 422) ->
 126;
yeccgoto('Reference', 24) ->
 407;
yeccgoto('Reference', 28) ->
 407;
yeccgoto('Reference', 410) ->
 407;
yeccgoto('Reference', 425) ->
 407;
yeccgoto('ReferencedType', 36) ->
 72;
yeccgoto('ReferencedType', 38) ->
 72;
yeccgoto('ReferencedType', 47) ->
 72;
yeccgoto('ReferencedType', 80) ->
 72;
yeccgoto('ReferencedType', 94) ->
 72;
yeccgoto('ReferencedType', 96) ->
 127;
yeccgoto('ReferencedType', 101) ->
 127;
yeccgoto('ReferencedType', 118) ->
 72;
yeccgoto('ReferencedType', 205) ->
 127;
yeccgoto('ReferencedType', 220) ->
 127;
yeccgoto('ReferencedType', 224) ->
 127;
yeccgoto('ReferencedType', 229) ->
 127;
yeccgoto('ReferencedType', 231) ->
 72;
yeccgoto('ReferencedType', 244) ->
 72;
yeccgoto('ReferencedType', 252) ->
 127;
yeccgoto('ReferencedType', 258) ->
 72;
yeccgoto('ReferencedType', 260) ->
 72;
yeccgoto('ReferencedType', 264) ->
 127;
yeccgoto('ReferencedType', 283) ->
 72;
yeccgoto('ReferencedType', 284) ->
 72;
yeccgoto('ReferencedType', 289) ->
 72;
yeccgoto('ReferencedType', 298) ->
 72;
yeccgoto('ReferencedType', 307) ->
 72;
yeccgoto('ReferencedType', 311) ->
 72;
yeccgoto('ReferencedType', 314) ->
 72;
yeccgoto('ReferencedType', 317) ->
 72;
yeccgoto('ReferencedType', 324) ->
 72;
yeccgoto('ReferencedType', 327) ->
 72;
yeccgoto('ReferencedType', 367) ->
 72;
yeccgoto('ReferencedType', 370) ->
 72;
yeccgoto('RequiredToken', 381) ->
 386;
yeccgoto('RequiredToken', 387) ->
 386;
yeccgoto('RequiredToken', 390) ->
 386;
yeccgoto('RestrictedCharacterStringValue', 96) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 101) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 125) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 138) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 148) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 149) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 166) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 178) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 205) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 210) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 211) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 220) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 224) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 229) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 241) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 244) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 252) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 258) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 260) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 264) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 301) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 370) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 402) ->
 128;
yeccgoto('RestrictedCharacterStringValue', 422) ->
 128;
yeccgoto('SAndSOfValue', 96) ->
 129;
yeccgoto('SAndSOfValue', 101) ->
 129;
yeccgoto('SAndSOfValue', 125) ->
 129;
yeccgoto('SAndSOfValue', 138) ->
 129;
yeccgoto('SAndSOfValue', 148) ->
 129;
yeccgoto('SAndSOfValue', 149) ->
 129;
yeccgoto('SAndSOfValue', 166) ->
 129;
yeccgoto('SAndSOfValue', 178) ->
 129;
yeccgoto('SAndSOfValue', 205) ->
 129;
yeccgoto('SAndSOfValue', 210) ->
 129;
yeccgoto('SAndSOfValue', 211) ->
 129;
yeccgoto('SAndSOfValue', 220) ->
 129;
yeccgoto('SAndSOfValue', 224) ->
 129;
yeccgoto('SAndSOfValue', 229) ->
 129;
yeccgoto('SAndSOfValue', 241) ->
 129;
yeccgoto('SAndSOfValue', 244) ->
 129;
yeccgoto('SAndSOfValue', 252) ->
 129;
yeccgoto('SAndSOfValue', 258) ->
 129;
yeccgoto('SAndSOfValue', 260) ->
 129;
yeccgoto('SAndSOfValue', 264) ->
 129;
yeccgoto('SAndSOfValue', 301) ->
 129;
yeccgoto('SAndSOfValue', 370) ->
 129;
yeccgoto('SAndSOfValue', 402) ->
 129;
yeccgoto('SAndSOfValue', 422) ->
 129;
yeccgoto('SelectionType', 36) ->
 75;
yeccgoto('SelectionType', 38) ->
 75;
yeccgoto('SelectionType', 47) ->
 75;
yeccgoto('SelectionType', 80) ->
 75;
yeccgoto('SelectionType', 94) ->
 75;
yeccgoto('SelectionType', 96) ->
 75;
yeccgoto('SelectionType', 101) ->
 75;
yeccgoto('SelectionType', 118) ->
 75;
yeccgoto('SelectionType', 205) ->
 75;
yeccgoto('SelectionType', 220) ->
 75;
yeccgoto('SelectionType', 224) ->
 75;
yeccgoto('SelectionType', 229) ->
 75;
yeccgoto('SelectionType', 231) ->
 75;
yeccgoto('SelectionType', 244) ->
 75;
yeccgoto('SelectionType', 252) ->
 75;
yeccgoto('SelectionType', 258) ->
 75;
yeccgoto('SelectionType', 260) ->
 75;
yeccgoto('SelectionType', 264) ->
 75;
yeccgoto('SelectionType', 283) ->
 75;
yeccgoto('SelectionType', 284) ->
 75;
yeccgoto('SelectionType', 289) ->
 75;
yeccgoto('SelectionType', 291) ->
 297;
yeccgoto('SelectionType', 298) ->
 75;
yeccgoto('SelectionType', 305) ->
 297;
yeccgoto('SelectionType', 307) ->
 75;
yeccgoto('SelectionType', 311) ->
 75;
yeccgoto('SelectionType', 314) ->
 75;
yeccgoto('SelectionType', 317) ->
 75;
yeccgoto('SelectionType', 319) ->
 297;
yeccgoto('SelectionType', 324) ->
 75;
yeccgoto('SelectionType', 327) ->
 75;
yeccgoto('SelectionType', 349) ->
 297;
yeccgoto('SelectionType', 367) ->
 75;
yeccgoto('SelectionType', 370) ->
 75;
yeccgoto('SequenceOfType', 36) ->
 76;
yeccgoto('SequenceOfType', 38) ->
 76;
yeccgoto('SequenceOfType', 47) ->
 76;
yeccgoto('SequenceOfType', 80) ->
 76;
yeccgoto('SequenceOfType', 94) ->
 76;
yeccgoto('SequenceOfType', 118) ->
 76;
yeccgoto('SequenceOfType', 231) ->
 76;
yeccgoto('SequenceOfType', 244) ->
 76;
yeccgoto('SequenceOfType', 258) ->
 76;
yeccgoto('SequenceOfType', 260) ->
 76;
yeccgoto('SequenceOfType', 283) ->
 76;
yeccgoto('SequenceOfType', 284) ->
 76;
yeccgoto('SequenceOfType', 289) ->
 76;
yeccgoto('SequenceOfType', 298) ->
 76;
yeccgoto('SequenceOfType', 307) ->
 76;
yeccgoto('SequenceOfType', 311) ->
 76;
yeccgoto('SequenceOfType', 314) ->
 76;
yeccgoto('SequenceOfType', 317) ->
 76;
yeccgoto('SequenceOfType', 324) ->
 76;
yeccgoto('SequenceOfType', 327) ->
 76;
yeccgoto('SequenceOfType', 367) ->
 76;
yeccgoto('SequenceOfType', 370) ->
 76;
yeccgoto('SequenceType', 36) ->
 77;
yeccgoto('SequenceType', 38) ->
 77;
yeccgoto('SequenceType', 47) ->
 77;
yeccgoto('SequenceType', 80) ->
 77;
yeccgoto('SequenceType', 94) ->
 77;
yeccgoto('SequenceType', 118) ->
 77;
yeccgoto('SequenceType', 231) ->
 77;
yeccgoto('SequenceType', 244) ->
 77;
yeccgoto('SequenceType', 258) ->
 77;
yeccgoto('SequenceType', 260) ->
 77;
yeccgoto('SequenceType', 283) ->
 77;
yeccgoto('SequenceType', 284) ->
 77;
yeccgoto('SequenceType', 289) ->
 77;
yeccgoto('SequenceType', 298) ->
 77;
yeccgoto('SequenceType', 307) ->
 77;
yeccgoto('SequenceType', 311) ->
 77;
yeccgoto('SequenceType', 314) ->
 77;
yeccgoto('SequenceType', 317) ->
 77;
yeccgoto('SequenceType', 324) ->
 77;
yeccgoto('SequenceType', 327) ->
 77;
yeccgoto('SequenceType', 367) ->
 77;
yeccgoto('SequenceType', 370) ->
 77;
yeccgoto('SetOfType', 36) ->
 78;
yeccgoto('SetOfType', 38) ->
 78;
yeccgoto('SetOfType', 47) ->
 78;
yeccgoto('SetOfType', 80) ->
 78;
yeccgoto('SetOfType', 94) ->
 78;
yeccgoto('SetOfType', 118) ->
 78;
yeccgoto('SetOfType', 231) ->
 78;
yeccgoto('SetOfType', 244) ->
 78;
yeccgoto('SetOfType', 258) ->
 78;
yeccgoto('SetOfType', 260) ->
 78;
yeccgoto('SetOfType', 283) ->
 78;
yeccgoto('SetOfType', 284) ->
 78;
yeccgoto('SetOfType', 289) ->
 78;
yeccgoto('SetOfType', 298) ->
 78;
yeccgoto('SetOfType', 307) ->
 78;
yeccgoto('SetOfType', 311) ->
 78;
yeccgoto('SetOfType', 314) ->
 78;
yeccgoto('SetOfType', 317) ->
 78;
yeccgoto('SetOfType', 324) ->
 78;
yeccgoto('SetOfType', 327) ->
 78;
yeccgoto('SetOfType', 367) ->
 78;
yeccgoto('SetOfType', 370) ->
 78;
yeccgoto('SetType', 36) ->
 79;
yeccgoto('SetType', 38) ->
 79;
yeccgoto('SetType', 47) ->
 79;
yeccgoto('SetType', 80) ->
 79;
yeccgoto('SetType', 94) ->
 79;
yeccgoto('SetType', 118) ->
 79;
yeccgoto('SetType', 231) ->
 79;
yeccgoto('SetType', 244) ->
 79;
yeccgoto('SetType', 258) ->
 79;
yeccgoto('SetType', 260) ->
 79;
yeccgoto('SetType', 283) ->
 79;
yeccgoto('SetType', 284) ->
 79;
yeccgoto('SetType', 289) ->
 79;
yeccgoto('SetType', 298) ->
 79;
yeccgoto('SetType', 307) ->
 79;
yeccgoto('SetType', 311) ->
 79;
yeccgoto('SetType', 314) ->
 79;
yeccgoto('SetType', 317) ->
 79;
yeccgoto('SetType', 324) ->
 79;
yeccgoto('SetType', 327) ->
 79;
yeccgoto('SetType', 367) ->
 79;
yeccgoto('SetType', 370) ->
 79;
yeccgoto('SignedNumber', 96) ->
 131;
yeccgoto('SignedNumber', 101) ->
 131;
yeccgoto('SignedNumber', 125) ->
 131;
yeccgoto('SignedNumber', 138) ->
 131;
yeccgoto('SignedNumber', 148) ->
 131;
yeccgoto('SignedNumber', 149) ->
 131;
yeccgoto('SignedNumber', 165) ->
 169;
yeccgoto('SignedNumber', 166) ->
 131;
yeccgoto('SignedNumber', 178) ->
 131;
yeccgoto('SignedNumber', 205) ->
 131;
yeccgoto('SignedNumber', 210) ->
 131;
yeccgoto('SignedNumber', 211) ->
 131;
yeccgoto('SignedNumber', 220) ->
 131;
yeccgoto('SignedNumber', 224) ->
 131;
yeccgoto('SignedNumber', 229) ->
 131;
yeccgoto('SignedNumber', 231) ->
 235;
yeccgoto('SignedNumber', 241) ->
 131;
yeccgoto('SignedNumber', 244) ->
 131;
yeccgoto('SignedNumber', 252) ->
 131;
yeccgoto('SignedNumber', 258) ->
 131;
yeccgoto('SignedNumber', 260) ->
 131;
yeccgoto('SignedNumber', 264) ->
 131;
yeccgoto('SignedNumber', 301) ->
 131;
yeccgoto('SignedNumber', 370) ->
 131;
yeccgoto('SignedNumber', 402) ->
 131;
yeccgoto('SignedNumber', 422) ->
 131;
yeccgoto('SubTypeElements', 96) ->
 132;
yeccgoto('SubTypeElements', 101) ->
 132;
yeccgoto('SubTypeElements', 205) ->
 132;
yeccgoto('SubTypeElements', 220) ->
 132;
yeccgoto('SubTypeElements', 224) ->
 132;
yeccgoto('SubTypeElements', 229) ->
 132;
yeccgoto('SubTypeElements', 252) ->
 132;
yeccgoto('SubTypeElements', 264) ->
 132;
yeccgoto('Symbol', 24) ->
 408;
yeccgoto('Symbol', 28) ->
 408;
yeccgoto('Symbol', 410) ->
 408;
yeccgoto('Symbol', 425) ->
 408;
yeccgoto('SymbolList', 24) ->
 430;
yeccgoto('SymbolList', 28) ->
 409;
yeccgoto('SymbolList', 410) ->
 409;
yeccgoto('SymbolList', 425) ->
 426;
yeccgoto('SymbolsFromModule', 28) ->
 410;
yeccgoto('SymbolsFromModule', 410) ->
 410;
yeccgoto('SymbolsFromModuleList', 28) ->
 411;
yeccgoto('SymbolsFromModuleList', 410) ->
 418;
yeccgoto('SyntaxList', 379) ->
 380;
yeccgoto('Tag', 36) ->
 80;
yeccgoto('Tag', 38) ->
 80;
yeccgoto('Tag', 47) ->
 80;
yeccgoto('Tag', 80) ->
 80;
yeccgoto('Tag', 94) ->
 80;
yeccgoto('Tag', 118) ->
 80;
yeccgoto('Tag', 231) ->
 80;
yeccgoto('Tag', 244) ->
 80;
yeccgoto('Tag', 258) ->
 80;
yeccgoto('Tag', 260) ->
 80;
yeccgoto('Tag', 283) ->
 80;
yeccgoto('Tag', 284) ->
 80;
yeccgoto('Tag', 289) ->
 80;
yeccgoto('Tag', 298) ->
 80;
yeccgoto('Tag', 307) ->
 80;
yeccgoto('Tag', 311) ->
 80;
yeccgoto('Tag', 314) ->
 80;
yeccgoto('Tag', 317) ->
 80;
yeccgoto('Tag', 324) ->
 80;
yeccgoto('Tag', 327) ->
 80;
yeccgoto('Tag', 367) ->
 80;
yeccgoto('Tag', 370) ->
 80;
yeccgoto('TagDefault', 15) ->
 19;
yeccgoto('TaggedType', 36) ->
 81;
yeccgoto('TaggedType', 38) ->
 81;
yeccgoto('TaggedType', 47) ->
 81;
yeccgoto('TaggedType', 80) ->
 81;
yeccgoto('TaggedType', 94) ->
 81;
yeccgoto('TaggedType', 118) ->
 81;
yeccgoto('TaggedType', 231) ->
 81;
yeccgoto('TaggedType', 244) ->
 81;
yeccgoto('TaggedType', 258) ->
 81;
yeccgoto('TaggedType', 260) ->
 81;
yeccgoto('TaggedType', 283) ->
 81;
yeccgoto('TaggedType', 284) ->
 81;
yeccgoto('TaggedType', 289) ->
 81;
yeccgoto('TaggedType', 298) ->
 81;
yeccgoto('TaggedType', 307) ->
 81;
yeccgoto('TaggedType', 311) ->
 81;
yeccgoto('TaggedType', 314) ->
 81;
yeccgoto('TaggedType', 317) ->
 81;
yeccgoto('TaggedType', 324) ->
 81;
yeccgoto('TaggedType', 327) ->
 81;
yeccgoto('TaggedType', 367) ->
 81;
yeccgoto('TaggedType', 370) ->
 81;
yeccgoto('TokenOrGroupSpec', 381) ->
 387;
yeccgoto('TokenOrGroupSpec', 387) ->
 387;
yeccgoto('TokenOrGroupSpec', 390) ->
 387;
yeccgoto('TokenOrGroupSpecs', 381) ->
 388;
yeccgoto('TokenOrGroupSpecs', 387) ->
 398;
yeccgoto('TokenOrGroupSpecs', 390) ->
 395;
yeccgoto('Tuple', 96) ->
 134;
yeccgoto('Tuple', 101) ->
 134;
yeccgoto('Tuple', 125) ->
 134;
yeccgoto('Tuple', 138) ->
 134;
yeccgoto('Tuple', 148) ->
 134;
yeccgoto('Tuple', 149) ->
 134;
yeccgoto('Tuple', 166) ->
 134;
yeccgoto('Tuple', 178) ->
 134;
yeccgoto('Tuple', 205) ->
 134;
yeccgoto('Tuple', 210) ->
 134;
yeccgoto('Tuple', 211) ->
 134;
yeccgoto('Tuple', 220) ->
 134;
yeccgoto('Tuple', 224) ->
 134;
yeccgoto('Tuple', 229) ->
 134;
yeccgoto('Tuple', 241) ->
 134;
yeccgoto('Tuple', 244) ->
 134;
yeccgoto('Tuple', 252) ->
 134;
yeccgoto('Tuple', 258) ->
 134;
yeccgoto('Tuple', 260) ->
 134;
yeccgoto('Tuple', 264) ->
 134;
yeccgoto('Tuple', 301) ->
 134;
yeccgoto('Tuple', 370) ->
 134;
yeccgoto('Tuple', 402) ->
 134;
yeccgoto('Tuple', 422) ->
 134;
yeccgoto('Type', 36) ->
 401;
yeccgoto('Type', 38) ->
 362;
yeccgoto('Type', 47) ->
 82;
yeccgoto('Type', 80) ->
 285;
yeccgoto('Type', 94) ->
 95;
yeccgoto('Type', 118) ->
 218;
yeccgoto('Type', 231) ->
 236;
yeccgoto('Type', 244) ->
 246;
yeccgoto('Type', 258) ->
 246;
yeccgoto('Type', 260) ->
 262;
yeccgoto('Type', 283) ->
 287;
yeccgoto('Type', 284) ->
 286;
yeccgoto('Type', 289) ->
 313;
yeccgoto('Type', 298) ->
 300;
yeccgoto('Type', 307) ->
 308;
yeccgoto('Type', 311) ->
 312;
yeccgoto('Type', 314) ->
 315;
yeccgoto('Type', 317) ->
 326;
yeccgoto('Type', 324) ->
 325;
yeccgoto('Type', 327) ->
 328;
yeccgoto('Type', 367) ->
 368;
yeccgoto('Type', 370) ->
 373;
yeccgoto('TypeAssignment', 29) ->
 34;
yeccgoto('TypeAssignment', 30) ->
 34;
yeccgoto('TypeConstraints', 184) ->
 187;
yeccgoto('TypeConstraints', 189) ->
 196;
yeccgoto('TypeConstraints', 198) ->
 199;
yeccgoto('TypeConstraints', 200) ->
 201;
yeccgoto('TypeFromObject', 36) ->
 83;
yeccgoto('TypeFromObject', 38) ->
 83;
yeccgoto('TypeFromObject', 47) ->
 83;
yeccgoto('TypeFromObject', 80) ->
 83;
yeccgoto('TypeFromObject', 94) ->
 83;
yeccgoto('TypeFromObject', 96) ->
 83;
yeccgoto('TypeFromObject', 101) ->
 83;
yeccgoto('TypeFromObject', 118) ->
 83;
yeccgoto('TypeFromObject', 205) ->
 83;
yeccgoto('TypeFromObject', 220) ->
 83;
yeccgoto('TypeFromObject', 224) ->
 83;
yeccgoto('TypeFromObject', 229) ->
 83;
yeccgoto('TypeFromObject', 231) ->
 83;
yeccgoto('TypeFromObject', 244) ->
 83;
yeccgoto('TypeFromObject', 252) ->
 83;
yeccgoto('TypeFromObject', 258) ->
 83;
yeccgoto('TypeFromObject', 260) ->
 83;
yeccgoto('TypeFromObject', 264) ->
 83;
yeccgoto('TypeFromObject', 283) ->
 83;
yeccgoto('TypeFromObject', 284) ->
 83;
yeccgoto('TypeFromObject', 289) ->
 83;
yeccgoto('TypeFromObject', 298) ->
 83;
yeccgoto('TypeFromObject', 307) ->
 83;
yeccgoto('TypeFromObject', 311) ->
 83;
yeccgoto('TypeFromObject', 314) ->
 83;
yeccgoto('TypeFromObject', 317) ->
 83;
yeccgoto('TypeFromObject', 324) ->
 83;
yeccgoto('TypeFromObject', 327) ->
 83;
yeccgoto('TypeFromObject', 367) ->
 83;
yeccgoto('TypeFromObject', 370) ->
 83;
yeccgoto('TypeWithConstraint', 36) ->
 84;
yeccgoto('TypeWithConstraint', 38) ->
 84;
yeccgoto('TypeWithConstraint', 47) ->
 84;
yeccgoto('TypeWithConstraint', 80) ->
 84;
yeccgoto('TypeWithConstraint', 94) ->
 84;
yeccgoto('TypeWithConstraint', 118) ->
 84;
yeccgoto('TypeWithConstraint', 231) ->
 84;
yeccgoto('TypeWithConstraint', 244) ->
 84;
yeccgoto('TypeWithConstraint', 258) ->
 84;
yeccgoto('TypeWithConstraint', 260) ->
 84;
yeccgoto('TypeWithConstraint', 283) ->
 84;
yeccgoto('TypeWithConstraint', 284) ->
 84;
yeccgoto('TypeWithConstraint', 289) ->
 84;
yeccgoto('TypeWithConstraint', 298) ->
 84;
yeccgoto('TypeWithConstraint', 307) ->
 84;
yeccgoto('TypeWithConstraint', 311) ->
 84;
yeccgoto('TypeWithConstraint', 314) ->
 84;
yeccgoto('TypeWithConstraint', 317) ->
 84;
yeccgoto('TypeWithConstraint', 324) ->
 84;
yeccgoto('TypeWithConstraint', 327) ->
 84;
yeccgoto('TypeWithConstraint', 367) ->
 84;
yeccgoto('TypeWithConstraint', 370) ->
 84;
yeccgoto('UElems', 96) ->
 135;
yeccgoto('UElems', 101) ->
 135;
yeccgoto('UElems', 229) ->
 135;
yeccgoto('UElems', 252) ->
 135;
yeccgoto('UElems', 264) ->
 135;
yeccgoto('UnionMark', 135) ->
 205;
yeccgoto('Unions', 96) ->
 136;
yeccgoto('Unions', 101) ->
 136;
yeccgoto('Unions', 229) ->
 136;
yeccgoto('Unions', 252) ->
 136;
yeccgoto('Unions', 264) ->
 136;
yeccgoto('UpperEndValue', 210) ->
 213;
yeccgoto('UpperEndValue', 211) ->
 216;
yeccgoto('UpperEndpoint', 210) ->
 214;
yeccgoto('UserDefinedConstraint', 96) ->
 137;
yeccgoto('UserDefinedConstraintParameter', 244) ->
 247;
yeccgoto('UserDefinedConstraintParameter', 258) ->
 247;
yeccgoto('UserDefinedConstraintParameters', 244) ->
 248;
yeccgoto('UserDefinedConstraintParameters', 258) ->
 259;
yeccgoto('Value', 96) ->
 138;
yeccgoto('Value', 101) ->
 138;
yeccgoto('Value', 125) ->
 149;
yeccgoto('Value', 138) ->
 149;
yeccgoto('Value', 148) ->
 149;
yeccgoto('Value', 149) ->
 149;
yeccgoto('Value', 166) ->
 167;
yeccgoto('Value', 178) ->
 149;
yeccgoto('Value', 205) ->
 138;
yeccgoto('Value', 210) ->
 215;
yeccgoto('Value', 211) ->
 215;
yeccgoto('Value', 220) ->
 138;
yeccgoto('Value', 224) ->
 138;
yeccgoto('Value', 229) ->
 138;
yeccgoto('Value', 241) ->
 242;
yeccgoto('Value', 252) ->
 138;
yeccgoto('Value', 264) ->
 138;
yeccgoto('Value', 301) ->
 303;
yeccgoto('Value', 402) ->
 403;
yeccgoto('Value', 422) ->
 149;
yeccgoto('ValueAssignment', 29) ->
 35;
yeccgoto('ValueAssignment', 30) ->
 35;
yeccgoto('ValueList', 96) ->
 139;
yeccgoto('ValueList', 101) ->
 139;
yeccgoto('ValueList', 125) ->
 209;
yeccgoto('ValueList', 138) ->
 179;
yeccgoto('ValueList', 148) ->
 150;
yeccgoto('ValueList', 149) ->
 179;
yeccgoto('ValueList', 178) ->
 180;
yeccgoto('ValueList', 205) ->
 139;
yeccgoto('ValueList', 220) ->
 139;
yeccgoto('ValueList', 224) ->
 139;
yeccgoto('ValueList', 229) ->
 139;
yeccgoto('ValueList', 252) ->
 255;
yeccgoto('ValueList', 264) ->
 139;
yeccgoto('ValueList', 422) ->
 423;
yeccgoto('ValueNotNull', 96) ->
 140;
yeccgoto('ValueNotNull', 101) ->
 140;
yeccgoto('ValueNotNull', 125) ->
 140;
yeccgoto('ValueNotNull', 138) ->
 140;
yeccgoto('ValueNotNull', 148) ->
 140;
yeccgoto('ValueNotNull', 149) ->
 140;
yeccgoto('ValueNotNull', 166) ->
 140;
yeccgoto('ValueNotNull', 178) ->
 140;
yeccgoto('ValueNotNull', 205) ->
 140;
yeccgoto('ValueNotNull', 210) ->
 140;
yeccgoto('ValueNotNull', 211) ->
 140;
yeccgoto('ValueNotNull', 220) ->
 140;
yeccgoto('ValueNotNull', 224) ->
 140;
yeccgoto('ValueNotNull', 229) ->
 140;
yeccgoto('ValueNotNull', 241) ->
 140;
yeccgoto('ValueNotNull', 244) ->
 249;
yeccgoto('ValueNotNull', 252) ->
 140;
yeccgoto('ValueNotNull', 258) ->
 249;
yeccgoto('ValueNotNull', 260) ->
 249;
yeccgoto('ValueNotNull', 264) ->
 140;
yeccgoto('ValueNotNull', 301) ->
 140;
yeccgoto('ValueNotNull', 370) ->
 374;
yeccgoto('ValueNotNull', 402) ->
 140;
yeccgoto('ValueNotNull', 422) ->
 140;
yeccgoto('ValueSet', 244) ->
 250;
yeccgoto('ValueSet', 258) ->
 250;
yeccgoto('ValueSet', 260) ->
 250;
yeccgoto('WithSyntaxSpec', 376) ->
 378;
yeccgoto('Word', 381) ->
 389;
yeccgoto('Word', 387) ->
 389;
yeccgoto('Word', 390) ->
 389;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


